@@ -31,10 +31,14 @@
 
 #include <glib.h>
 #include <request.h>
+#include <openssl/sha.h>
+
+#include "telegram-purple.h"
+#include "msglog.h"
+#include "tgp-2prpl.h"
+#include "tgp-structs.h"
+#include "lodepng/lodepng.h"
 
-#include <telegram-purple.h>
-#include <msglog.h>
-#include <tgp-2prpl.h>
 
 #define DC_SERIALIZED_MAGIC 0x868aa81d
 #define STATE_FILE_MAGIC 0x28949a93
@@ -163,6 +167,15 @@ void read_dc (struct tgl_state *TLS, int auth_file_fd, int id, unsigned ver) {
   bl_do_dc_signed (TLS, id);
 }
 
+int error_if_val_false (struct tgl_state *TLS, int val, const char *msg) {
+  if (!val) {
+    connection_data *conn = TLS->ev_base;
+    purple_connection_error_reason (conn->gc, PURPLE_CONNECTION_ERROR_AUTHENTICATION_FAILED, msg);
+    return 1;
+  }
+  return 0;
+}
+
 void empty_auth_file (struct tgl_state *TLS) {
   if (TLS->test_mode) {
     bl_do_dc_option (TLS, 1, 0, "", strlen (TG_SERVER_TEST_1), TG_SERVER_TEST_1, 443);
@@ -223,10 +236,141 @@ void read_auth_file (struct tgl_state *TLS) {
   close (auth_file_fd);
 }
 
+
+void write_secret_chat (tgl_peer_t *_P, void *extra) {
+  struct tgl_secret_chat *P = (void *)_P;
+  if (tgl_get_peer_type (P->id) != TGL_PEER_ENCR_CHAT) { return; }
+  if (P->state != sc_ok) { return; }
+  int *a = extra;
+  int fd = a[0];
+  a[1] ++;
+  
+  int id = tgl_get_peer_id (P->id);
+  assert (write (fd, &id, 4) == 4);
+  //assert (write (fd, &P->flags, 4) == 4);
+  int l = strlen (P->print_name);
+  assert (write (fd, &l, 4) == 4);
+  assert (write (fd, P->print_name, l) == l);
+  assert (write (fd, &P->user_id, 4) == 4);
+  assert (write (fd, &P->admin_id, 4) == 4);
+  assert (write (fd, &P->date, 4) == 4);
+  assert (write (fd, &P->ttl, 4) == 4);
+  assert (write (fd, &P->layer, 4) == 4);
+  assert (write (fd, &P->access_hash, 8) == 8);
+  assert (write (fd, &P->state, 4) == 4);
+  assert (write (fd, &P->key_fingerprint, 8) == 8);
+  assert (write (fd, &P->key, 256) == 256);
+  assert (write (fd, &P->first_key_sha, 20) == 20);
+  assert (write (fd, &P->in_seq_no, 4) == 4);
+  assert (write (fd, &P->last_in_seq_no, 4) == 4);
+  assert (write (fd, &P->out_seq_no, 4) == 4);
+}
+
+void write_secret_chat_file (struct tgl_state *TLS) {
+  char *name = 0;
+  if (asprintf (&name, "%s/%s", TLS->base_path, "secret") < 0) {
+    return;
+  }
+  int secret_chat_fd = open (name, O_CREAT | O_RDWR, 0600);
+  free (name);
+  assert (secret_chat_fd >= 0);
+  int x = SECRET_CHAT_FILE_MAGIC;
+  assert (write (secret_chat_fd, &x, 4) == 4);
+  x = 2;
+  assert (write (secret_chat_fd, &x, 4) == 4); // version
+  assert (write (secret_chat_fd, &x, 4) == 4); // num
+  
+  int y[2];
+  y[0] = secret_chat_fd;
+  y[1] = 0;
+  
+  tgl_peer_iterator_ex (TLS, write_secret_chat, y);
+  
+  lseek (secret_chat_fd, 8, SEEK_SET);
+  assert (write (secret_chat_fd, &y[1], 4) == 4);
+  close (secret_chat_fd);
+}
+
+void read_secret_chat (struct tgl_state *TLS, int fd, int v) {
+  int id, l, user_id, admin_id, date, ttl, layer, state;
+  long long access_hash, key_fingerprint;
+  static char s[1000];
+  static unsigned char key[256];
+  static unsigned char sha[20];
+  assert (read (fd, &id, 4) == 4);
+  //assert (read (fd, &flags, 4) == 4);
+  assert (read (fd, &l, 4) == 4);
+  assert (l > 0 && l < 1000);
+  assert (read (fd, s, l) == l);
+  assert (read (fd, &user_id, 4) == 4);
+  assert (read (fd, &admin_id, 4) == 4);
+  assert (read (fd, &date, 4) == 4);
+  assert (read (fd, &ttl, 4) == 4);
+  assert (read (fd, &layer, 4) == 4);
+  assert (read (fd, &access_hash, 8) == 8);
+  assert (read (fd, &state, 4) == 4);
+  assert (read (fd, &key_fingerprint, 8) == 8);
+  assert (read (fd, &key, 256) == 256);
+  if (v >= 2) {
+    assert (read (fd, sha, 20) == 20);
+  }
+  int in_seq_no = 0, out_seq_no = 0, last_in_seq_no = 0;
+  if (v >= 1) {
+    assert (read (fd, &in_seq_no, 4) == 4);
+    assert (read (fd, &last_in_seq_no, 4) == 4);
+    assert (read (fd, &out_seq_no, 4) == 4);
+  }
+  
+  bl_do_encr_chat_create (TLS, id, user_id, admin_id, s, l);
+  struct tgl_secret_chat  *P = (void *)tgl_peer_get (TLS, TGL_MK_ENCR_CHAT (id));
+  assert (P && (P->flags & FLAG_CREATED));
+  bl_do_encr_chat_set_date (TLS, P, date);
+  bl_do_encr_chat_set_ttl (TLS, P, ttl);
+  bl_do_encr_chat_set_layer (TLS ,P, layer);
+  bl_do_encr_chat_set_state (TLS, P, state);
+  bl_do_encr_chat_set_key (TLS, P, key, key_fingerprint);
+  if (v >= 2) {
+    bl_do_encr_chat_set_sha (TLS, P, sha);
+  } else {
+    SHA1 ((void *)key, 256, sha);
+    bl_do_encr_chat_set_sha (TLS, P, sha);
+  }
+  if (v >= 1) {
+    bl_do_encr_chat_set_seq (TLS, P, in_seq_no, last_in_seq_no, out_seq_no);
+  }
+  bl_do_encr_chat_set_access_hash (TLS, P, access_hash);
+}
+
+void read_secret_chat_file (struct tgl_state *TLS) {
+  char *name = 0;
+  if (asprintf (&name, "%s/%s", TLS->base_path, "secret") < 0) {
+    return;
+  }
+  
+  int secret_chat_fd = open (name, O_RDWR, 0600);
+  free (name);
+  
+  if (secret_chat_fd < 0) { return; }
+  
+  int x;
+  if (read (secret_chat_fd, &x, 4) < 4) { close (secret_chat_fd); return; }
+  if (x != SECRET_CHAT_FILE_MAGIC) { close (secret_chat_fd); return; }
+  int v = 0;
+  assert (read (secret_chat_fd, &v, 4) == 4);
+  assert (v == 0 || v == 1 || v == 2); // version
+  assert (read (secret_chat_fd, &x, 4) == 4);
+  assert (x >= 0);
+  while (x -- > 0) {
+    read_secret_chat (TLS, secret_chat_fd, v);
+  }
+  close (secret_chat_fd);
+}
+
 void telegram_export_authorization (struct tgl_state *TLS);
 void export_auth_callback (struct tgl_state *TLS, void *extra, int success) {
-  assert (success);
-  telegram_export_authorization (TLS);
+  if (!error_if_val_false(TLS, success, "Authentication Export failed.")) {
+    telegram_export_authorization (TLS);
+  }
 }
 
 void telegram_export_authorization (struct tgl_state *TLS) {
@@ -259,23 +403,23 @@ static void code_auth_receive_result (struct tgl_state *TLS, void *extra, int su
   }
 }
 
-static void request_code_entered (gpointer data, const gchar *code) {
+void request_code_entered (gpointer data, const gchar *code) {
   struct tgl_state *TLS = data;
-  telegram_conn *conn = TLS->ev_base;
+  connection_data *conn = TLS->ev_base;
   char const *username = purple_account_get_username(conn->pa);
   tgl_do_send_code_result (TLS, username, conn->hash, code, code_receive_result, 0) ;
 }
 
 static void request_code_canceled (gpointer data) {
   struct tgl_state *TLS = data;
-  telegram_conn *conn = TLS->ev_base;
+  connection_data *conn = TLS->ev_base;
 
   purple_connection_error_reason(conn->gc,
       PURPLE_CONNECTION_ERROR_AUTHENTICATION_FAILED, "registration canceled");
 }
 
 static void request_name_code_entered (PurpleConnection* gc, PurpleRequestFields* fields) {
-  telegram_conn *conn = purple_connection_get_protocol_data(gc);
+  connection_data *conn = purple_connection_get_protocol_data(gc);
   struct tgl_state *TLS = conn->TLS;
   char const *username = purple_account_get_username(conn->pa);
   
@@ -292,30 +436,29 @@ static void request_name_code_entered (PurpleConnection* gc, PurpleRequestFields
 
 static void request_code (struct tgl_state *TLS) {
   debug ("Client is not registered, registering...\n");
-  telegram_conn *conn = TLS->ev_base;
+  connection_data *conn = TLS->ev_base;
   int compat = purple_account_get_bool (tg_get_acc(TLS), "compat-verification", 0);
   
   if (compat || ! purple_request_input (conn->gc, "Telegram Code", "Enter Telegram Code",
         "Telegram wants to verify your identity, please enter the code, that you have received via SMS.",
         NULL, 0, 0, "code", "OK", G_CALLBACK(request_code_entered), "Cancel",
         G_CALLBACK(request_code_canceled), conn->pa, NULL, NULL, TLS)) {
-      const char *sms  = purple_account_get_string(tg_get_acc(TLS), "code", "");
-      if (*sms == '\0') {
-        const char *error = "Could not prompt for sms code, please set SMS code in account settings and reconnect.";
-        purple_connection_error_reason(conn->gc, PURPLE_CONNECTION_ERROR_AUTHENTICATION_FAILED, error);
-        purple_notify_error(_telegram_protocol, "Enter SMS code", "Enter SMS code", error);
-        return;
-      }
-      
-      request_code_entered (TLS, sms);
-      purple_account_set_string(tg_get_acc(TLS), "code", "");
+
+    // purple request API is not available, so we create a new conversation (the Telegram system
+    // account "7770000") to prompt the user for the code
+          
+    conn->in_fallback_chat = 1;
+    purple_connection_set_state (conn->gc, PURPLE_CONNECTED);
+    PurpleConversation *conv = purple_conversation_new (PURPLE_CONV_TYPE_IM, conn->pa, "777000");
+    purple_conversation_write (conv, "777000", "What is your SMS verification code?",
+          PURPLE_MESSAGE_RECV | PURPLE_MESSAGE_SYSTEM, 0);
   }
 }
 
 static void request_name_and_code (struct tgl_state *TLS) {
   debug ("Phone is not registered, registering...\n");
 
-  telegram_conn *conn = TLS->ev_base;
+  connection_data *conn = TLS->ev_base;
 
   PurpleRequestFields* fields = purple_request_fields_new();
   PurpleRequestField* field = 0;
@@ -342,14 +485,14 @@ static void request_name_and_code (struct tgl_state *TLS) {
 }
 
 static void sign_in_callback (struct tgl_state *TLS, void *extra, int success, int registered, const char *mhash) {
-  assert (success); // TODO proper error handle
-  telegram_conn *conn = TLS->ev_base;
-  conn->hash = strdup (mhash);
-
-  if (registered) {
-    request_code (TLS);
-  } else {
-    request_name_and_code (TLS);
+  connection_data *conn = TLS->ev_base;
+  if (!error_if_val_false (TLS, success, "Invalid or non-existing phone number.")) {
+    conn->hash = strdup (mhash);
+    if (registered) {
+      request_code (TLS);
+    } else {
+      request_name_and_code (TLS);
+    }
   }
 }
 
@@ -358,7 +501,7 @@ static void telegram_send_sms (struct tgl_state *TLS) {
     telegram_export_authorization (TLS);
     return;
   }
-  telegram_conn *conn = TLS->ev_base;
+  connection_data *conn = TLS->ev_base;
   char const *username = purple_account_get_username(conn->pa);
   tgl_do_send_code (TLS, username, sign_in_callback, 0);
 }
@@ -386,6 +529,7 @@ static int check_all_authorized (gpointer arg) {
 void telegram_login (struct tgl_state *TLS) {    
   read_auth_file (TLS);
   read_state_file (TLS);
+  read_secret_chat_file (TLS);
   if (all_authorized (TLS)) {
     telegram_send_sms (TLS);
     return;
@@ -395,7 +539,7 @@ void telegram_login (struct tgl_state *TLS) {
 
 PurpleConversation *chat_show (PurpleConnection *gc, int id) {
   debug ("show chat");
-  telegram_conn *conn = purple_connection_get_protocol_data(gc);
+  connection_data *conn = purple_connection_get_protocol_data(gc);
   
   PurpleConversation *convo = purple_find_chat(gc, id);
   if (! convo) {
@@ -411,17 +555,20 @@ PurpleConversation *chat_show (PurpleConnection *gc, int id) {
 }
 
 int chat_add_message (struct tgl_state *TLS, struct tgl_message *M, char *text) {
-  telegram_conn *conn = TLS->ev_base;
+  connection_data *conn = TLS->ev_base;
   
   if (chat_show (conn->gc, tgl_get_peer_id (M->to_id))) {
     p2tgl_got_chat_in(TLS, M->to_id, M->from_id, text ? text : M->message,
                       M->service ? PURPLE_MESSAGE_SYSTEM : PURPLE_MESSAGE_RECV, M->date);
+    
+    pending_reads_add (conn->pending_reads, M->to_id);
+    if (p2tgl_status_is_present(purple_account_get_active_status(conn->pa))) {
+      pending_reads_send_all (conn->pending_reads, conn->TLS);
+    }
     return 1;
   } else {
     // add message once the chat was initialised
-    struct message_text *mt = malloc (sizeof (*mt));
-    mt->M = M;
-    mt->text = text ? g_strdup (text) : text;
+    struct message_text *mt = message_text_init (M, text);
     g_queue_push_tail (conn->new_messages, mt);
     return 0;
   }
@@ -441,3 +588,57 @@ void chat_add_all_users (PurpleConversation *pc, struct tgl_chat *chat) {
     p2tgl_conv_add_user(pc, *uid, NULL, flags, 0);
   }
 }
+
+/**
+ * This function generates a png image to visualize the sha1 key from an encrypted chat.
+ */
+int generate_ident_icon (struct tgl_state *TLS, unsigned char* sha1_key)
+{
+  int colors[4] = {
+    0xffffff,
+    0xd5e6f3,
+    0x2d5775,
+    0x2f99c9
+  };
+  unsigned img_size = 160;
+  unsigned char* image = (unsigned char*)malloc (img_size * img_size * 4);
+  assert (image);
+  unsigned x, y, i, j, idx = 0;
+  int bitpointer = 0;
+  for (y = 0; y < 8; y++)
+  {
+    unsigned offset_y = y * img_size * 4 * (img_size / 8);
+    for (x = 0; x < 8; x++)
+    {
+      int offset = bitpointer / 8;
+      int shiftOffset = bitpointer % 8;
+      int val = sha1_key[offset + 3] << 24 | sha1_key[offset + 2] << 16 | sha1_key[offset + 1] << 8 | sha1_key[offset];
+      idx = abs((val >> shiftOffset) & 3) % 4;
+      bitpointer += 2;
+      unsigned offset_x = x * 4 * (img_size / 8);
+      for (i = 0; i < img_size / 8; i++)
+      {
+        unsigned off_y = offset_y + i * img_size * 4;
+        for (j = 0; j < img_size / 8; j++)
+        {
+          unsigned off_x = offset_x + j * 4;
+          image[off_y + off_x + 0] = (colors[idx] >> 16) & 0xFF;
+          image[off_y + off_x + 1] = (colors[idx] >> 8) & 0xFF;
+          image[off_y + off_x + 2] = colors[idx] & 0xFF;
+          image[off_y + off_x + 3] = 0xFF;
+        }
+      }
+    }
+  }
+  unsigned char* png;
+  size_t pngsize;
+  unsigned error = lodepng_encode32(&png, &pngsize, image, img_size, img_size);
+  int imgStoreId = -1;
+  if(!error)
+  {
+    imgStoreId = purple_imgstore_add_with_id (png, pngsize, NULL);
+    used_images_add ((connection_data*)TLS->ev_base, imgStoreId);
+  }
+  g_free(image);
+  return imgStoreId;
+}
