@@ -51,6 +51,9 @@
 #include "request.h"
 
 #include <tgl.h>
+#include <tgl-binlog.h>
+#include <tools.h>
+#include "tgp-structs.h"
 #include "tgp-2prpl.h"
 #include "tgp-net.h"
 #include "tgp-timers.h"
@@ -58,16 +61,17 @@
 #include "telegram-purple.h"
 #include "msglog.h"
 
+#define _(m) m
+
 PurplePlugin *_telegram_protocol = NULL;
 PurpleGroup *tggroup;
 const char *config_dir = ".telegram-purple";
 const char *pk_path = "/etc/telegram-purple/server.pub";
-
 void tgprpl_login_on_connected();
-void on_user_get_info (struct tgl_state *TLS, void *show_info, int success, struct tgl_user *U);
+void on_user_get_info (struct tgl_state *TLS, void *info_data, int success, struct tgl_user *U);
 
-static telegram_conn *get_conn_from_buddy (PurpleBuddy *buddy) {
-  telegram_conn *c = purple_connection_get_protocol_data (
+static connection_data *get_conn_from_buddy (PurpleBuddy *buddy) {
+  connection_data *c = purple_connection_get_protocol_data (
       purple_account_get_connection (purple_buddy_get_account (buddy)));
   return c;
 }
@@ -82,13 +86,13 @@ static char *format_status (struct tgl_user_status *status) {
 }
 
 static char *format_img_full (int imgstore) {
-  return g_strdup_printf ("<br><img id=\"%u\">", imgstore);
-}
-
-static char *format_img_thumb (int imgstore, char *filename) __attribute__ ((unused));
-static char *format_img_thumb (int imgstore, char *filename) {
-  return g_strdup_printf("<a href=\"%s\"><img id=\"%u\"></a><br/><a href=\"%s\">%s</a>",
-           filename, imgstore, filename, filename);
+  const char *br = "<br>";
+  
+  // <br>'s look ugly in Adium, but no <br> will look ugly in Pidgin
+#ifdef __ADIUM_
+  br = "";
+#endif
+  return g_strdup_printf ("%s<img id=\"%u\">",br, imgstore);
 }
 
 static char *format_service_msg (struct tgl_state *TLS, struct tgl_message *M)
@@ -153,7 +157,6 @@ static char *format_service_msg (struct tgl_state *TLS, struct tgl_message *M)
     case tgl_message_action_notify_layer:
       txt_action = g_strdup_printf ("updated layer to %d", M->action.layer);
       break;
-      /*
     case tgl_message_action_request_key:
       txt_action = g_strdup_printf ("Request rekey #%016llx\n", M->action.exchange_id);
       break;
@@ -166,7 +169,6 @@ static char *format_service_msg (struct tgl_state *TLS, struct tgl_message *M)
     case tgl_message_action_abort_key:
       txt_action = g_strdup_printf ("Abort rekey #%016llx\n", M->action.exchange_id);
       break;
-      */
     default:
       txt_action = NULL;
       break;
@@ -180,14 +182,114 @@ static char *format_service_msg (struct tgl_state *TLS, struct tgl_message *M)
   return txt;
 }
 
+static char *format_print_name (struct tgl_state *TLS, tgl_peer_id_t id, const char *a1, const char *a2, const char *a3, const char *a4) {
+  const char *d[4];
+  d[0] = a1; d[1] = a2; d[2] = a3; d[3] = a4;
+  static char buf[10000];
+  buf[0] = 0;
+  int i;
+  int p = 0;
+  for (i = 0; i < 4; i++) {
+    if (d[i] && strlen (d[i])) {
+      p += tgl_snprintf (buf + p, 9999 - p, "%s%s", p ? " " : "", d[i]);
+      assert (p < 9990);
+    }
+  }
+  char *s = buf;
+  while (*s) {
+    if (*s == '\n') { *s = ' '; }
+    if (*s == '#') { *s = '@'; }
+    s++;
+  }
+  s = buf;
+  int fl = strlen (s);
+  int cc = 0;
+  while (1) {
+    tgl_peer_t *P = tgl_peer_get_by_name (TLS, s);
+    if (!P || !tgl_cmp_peer_id (P->id, id)) {
+      break;
+    }
+    cc ++;
+    assert (cc <= 9999);
+    tgl_snprintf (s + fl, 9999 - fl, " #%d", cc);
+  }
+  return tgl_strdup (s);
+}
+
+static char *format_document_desc (char *type, char *caption, gint64 size) {
+  char *s = g_format_size_for_display (size);
+  char *msg = g_strdup_printf ("[%s] %s %s", type, caption, s);
+  g_free (s);
+  return msg;
+}
+
+static char *format_message (struct tgl_message *M) {
+
+  switch (M->media.type) {
+    case tgl_message_media_audio:
+      return format_document_desc("AUDIO", "", M->media.audio.size);
+      break;
+    case tgl_message_media_audio_encr:
+      return format_document_desc("AUDIO", "", M->media.encr_audio.size);
+      break;
+    case tgl_message_media_document:
+      return format_document_desc("DOCUMENT", M->media.document.caption, M->media.document.size);
+      break;
+    case tgl_message_media_document_encr:
+      return format_document_desc("DOCUMENT", M->media.encr_document.file_name, M->media.encr_document.size);
+      break;
+    case tgl_message_media_video:
+      return format_document_desc("VIDEO", M->media.video.caption, M->media.video.size);
+      break;
+    case tgl_message_media_video_encr:
+      return format_document_desc("VIDEO", "", M->media.encr_video.size);
+      break;
+    case tgl_message_media_photo_encr:
+      return format_document_desc("PHOTO", "", M->media.encr_photo.size);
+      break;
+    case tgl_message_media_contact:
+      return g_strdup ("[CONTACT]");
+      break;
+    default:
+      if (M->message && *M->message != 0) {
+        return purple_markup_escape_text (M->message, strlen (M->message));
+      }
+      return g_strdup("");
+      break;
+  }
+}
+
+
+static void tgl_do_send_unescape_message (struct tgl_state *TLS, const char *message, tgl_peer_id_t to)
+{
+  gchar *raw = purple_unescape_html(message);
+  tgl_do_send_message (TLS, to, raw, (int)strlen (raw), 0, 0);
+  g_free(raw);
+}
+
+static tgl_peer_t *find_peer_by_name (struct tgl_state *TLS, const char *who) {
+  tgl_peer_t *peer = tgl_peer_get (TLS, TGL_MK_USER(atoi (who)));
+  if (peer) { return peer; }
+  peer = tgl_peer_get (TLS, TGL_MK_CHAT(atoi(who)));
+  if (peer) { return peer; }
+  peer = tgl_peer_get (TLS, TGL_MK_ENCR_CHAT(atoi(who)));
+  if (peer) { return peer; }
+  return NULL;
+}
+
 static int our_msg (struct tgl_state *TLS, struct tgl_message *M) {
   //return tgl_get_peer_id(M->from_id) == TLS->our_id;
+  //return M->out;
+  return (M->flags & FLAG_SESSION_OUTBOUND) != 0;
+}
+
+static int out_msg (struct tgl_state *TLS, struct tgl_message *M) {
   return M->out;
 }
 
 static gboolean queries_timerfunc (gpointer data) {
   debug ("queries_timerfunc()\n");
-  telegram_conn *conn = data;
+  connection_data *conn = data;
   
   if (conn->updated) {
     conn->updated = 0;
@@ -196,6 +298,15 @@ static gboolean queries_timerfunc (gpointer data) {
   return 1;
 }
 
+static void start_secret_chat (PurpleBlistNode *node, gpointer data) {
+  PurpleBuddy *buddy = data;
+  connection_data *conn = purple_connection_get_protocol_data (
+                           purple_account_get_connection (purple_buddy_get_account(buddy)));
+  
+  tgl_do_create_secret_chat(conn->TLS, TGL_MK_USER(atoi (purple_buddy_get_name (buddy))),
+                            0, 0);
+}
+
 static void on_update_user_name (struct tgl_state *TLS, tgl_peer_t *user) __attribute__ ((unused));
 static void on_update_user_name (struct tgl_state *TLS, tgl_peer_t *user) {
   p2tgl_got_alias(TLS, user->id, p2tgl_strdup_alias(user));
@@ -218,6 +329,7 @@ static void on_update_new_user_status (struct tgl_state *TLS, void *peer) {
 static void update_message_received (struct tgl_state *TLS, struct tgl_message *M);
 static void update_user_handler (struct tgl_state *TLS, struct tgl_user *U, unsigned flags);
 static void update_chat_handler (struct tgl_state *TLS, struct tgl_chat *C, unsigned flags);
+static void update_secret_chat_handler (struct tgl_state *TLS, struct tgl_secret_chat *C, unsigned flags);
 static void update_user_typing (struct tgl_state *TLS, struct tgl_user *U, enum tgl_typing_status status);
 struct tgl_update_callback tgp_callback = {
   .logprintf = debug,
@@ -225,50 +337,49 @@ struct tgl_update_callback tgp_callback = {
   .msg_receive = update_message_received,
   .user_update = update_user_handler,
   .chat_update = update_chat_handler,
-  .type_notification = update_user_typing
+  .secret_chat_update = update_secret_chat_handler,
+  .type_notification = update_user_typing,
+  .create_print_name = format_print_name
 };
 
 void on_message_load_photo (struct tgl_state *TLS, void *extra, int success, char *filename) {
+  connection_data *conn = TLS->ev_base;
+  
   gchar *data = NULL;
   size_t len;
   GError *err = NULL;
   g_file_get_contents (filename, &data, &len, &err);
   int imgStoreId = purple_imgstore_add_with_id (g_memdup(data, (guint)len), len, NULL);
-    
-  char *image = format_img_full(imgStoreId);
+  used_images_add (conn, imgStoreId);
+  
+  char *image = format_img_full (imgStoreId);
   struct tgl_message *M = extra;
   switch (tgl_get_peer_type (M->to_id)) {
     case TGL_PEER_CHAT:
       debug ("PEER_CHAT\n");
-      if (! our_msg(TLS, M)) {
+      if (!our_msg(TLS, M)) {
         chat_add_message (TLS, M, image);
       }
       break;
       
     case TGL_PEER_USER:
       debug ("PEER_USER\n");
-      if (our_msg(TLS, M)) {
+      if (out_msg(TLS, M)) {
         p2tgl_got_im (TLS, M->to_id, image, PURPLE_MESSAGE_SEND | PURPLE_MESSAGE_IMAGES, M->date);
       } else {
         p2tgl_got_im (TLS, M->from_id, image, PURPLE_MESSAGE_RECV | PURPLE_MESSAGE_IMAGES, M->date);
       }
       break;
-      
-    case TGL_PEER_ENCR_CHAT:
-      break;
-      
-    case TGL_PEER_GEO_CHAT:
-      break;
   }
  
   g_free (image);
-  telegram_conn *conn = TLS->ev_base;
+  conn = TLS->ev_base;
   conn->updated = 1;
 }
 
 static void update_message_received (struct tgl_state *TLS, struct tgl_message *M) {
   debug ("received message\n");
-  telegram_conn *conn = TLS->ev_base;
+  connection_data *conn = TLS->ev_base;
   conn->updated = 1;
 
   if (M->service) {
@@ -303,33 +414,38 @@ static void update_message_received (struct tgl_state *TLS, struct tgl_message *
     return;
   }
 
-  if (!M->message) {
+  if (!M->message || our_msg(TLS, M)) {
     return;
   }
 
-  char *text = purple_markup_escape_text (M->message, strlen (M->message));
+  char *text = format_message(M);
   switch (tgl_get_peer_type (M->to_id)) {
     case TGL_PEER_CHAT:
-      debug ("PEER_CHAT\n");
-      if (! our_msg(TLS, M)) {
-        chat_add_message (TLS, M, text);
-      }
-      break;
-      
-    case TGL_PEER_USER:
-      debug ("PEER_USER\n");
-      
-      // p2tgl_got_im (TLS, M->to_id, text, PURPLE_MESSAGE_SEND, M->date);
-      // :TODO: figure out how to add messages from different devices to history
-      if (! our_msg(TLS, M)) {
-        p2tgl_got_im (TLS, M->from_id, text, PURPLE_MESSAGE_RECV, M->date);
-      }
+      chat_add_message (TLS, M, text);
       break;
       
     case TGL_PEER_ENCR_CHAT:
+        p2tgl_got_im (TLS, M->to_id, text, PURPLE_MESSAGE_RECV, M->date);
+        
+        pending_reads_add (conn->pending_reads, M->to_id);
+        if (p2tgl_status_is_present (purple_account_get_active_status(conn->pa))) {
+          pending_reads_send_all (conn->pending_reads, conn->TLS);
+        }
       break;
       
-    case TGL_PEER_GEO_CHAT:
+    case TGL_PEER_USER:
+      
+      // :TODO: figure out how to add messages from different devices to history
+      if (out_msg(TLS, M)) {
+        p2tgl_got_im (TLS, M->to_id, text, PURPLE_MESSAGE_SEND, M->date);
+      } else {
+        p2tgl_got_im (TLS, M->from_id, text, PURPLE_MESSAGE_RECV, M->date);
+        
+        pending_reads_add (conn->pending_reads, M->from_id);
+        if (p2tgl_status_is_present (purple_account_get_active_status(conn->pa))) {
+          pending_reads_send_all (conn->pending_reads, conn->TLS);
+        }
+      }
       break;
   }
   
@@ -337,7 +453,6 @@ static void update_message_received (struct tgl_state *TLS, struct tgl_message *
 }
 
 static void update_user_handler (struct tgl_state *TLS, struct tgl_user *user, unsigned flags) {
-//  if (!(flags & TGL_UPDATE_CREATED)) { return; }
   if (TLS->our_id == tgl_get_peer_id (user->id)) {
     if (flags & TGL_UPDATE_NAME) {
       p2tgl_connection_set_display_name (TLS, (tgl_peer_t *)user);
@@ -353,11 +468,11 @@ static void update_user_handler (struct tgl_state *TLS, struct tgl_user *user, u
       p2tgl_prpl_got_user_status (TLS, user->id, &user->status);
       p2tgl_buddy_update (TLS, (tgl_peer_t *)user, flags);
     }
-    if (flags & (TGL_UPDATE_NAME | TGL_UPDATE_REAL_NAME | TGL_UPDATE_USERNAME) && buddy) {
-      p2tgl_blist_alias_buddy (buddy, user);
-    }
     if (flags & TGL_UPDATE_PHOTO) {
-      tgl_do_get_user_info (TLS, user->id, 0, on_user_get_info, 0);
+      get_user_info_data* info_data = malloc (sizeof(get_user_info_data));
+      info_data->show_info = 0;
+      info_data->peer = 0;
+      tgl_do_get_user_info (TLS, user->id, 0, on_user_get_info, info_data);
     }
     if (flags & TGL_UPDATE_DELETED && buddy) {
       purple_blist_remove_buddy (buddy);
@@ -365,8 +480,89 @@ static void update_user_handler (struct tgl_state *TLS, struct tgl_user *user, u
   }
 }
 
+static void write_secret_chat_cb (struct tgl_state *TLS, void *extra, int success, struct tgl_secret_chat *E) {
+  debug ("update_secret_chat_handle success=%d", success);
+  write_secret_chat_file (TLS);
+}
+
+struct accept_secret_chat_data {
+  struct tgl_state *TLS;
+  struct tgl_secret_chat *U;
+};
+
+static void accept_secret_chat_cb (gpointer _data, const gchar *code) {
+  struct accept_secret_chat_data *data = _data;
+  
+  tgl_do_accept_encr_chat_request (data->TLS, data->U, write_secret_chat_cb, 0);
+  
+  g_free (data);
+}
+
+static void decline_secret_chat_cb (gpointer _data, const gchar *code) {
+  struct accept_secret_chat_data *data = _data;
+  
+  /* TODO: implement the api call to cancel secret chats, see tgprpl_remove_buddy */
+  bl_do_encr_chat_delete (data->TLS, data->U);
+  purple_blist_remove_buddy (p2tgl_buddy_find(data->TLS, data->U->id));
+  
+  g_free (data);
+}
+
+static void update_secret_chat_handler (struct tgl_state *TLS, struct tgl_secret_chat *U, unsigned flags) {
+  debug ("secret-chat-state: %d", U->state);
+  
+  if (flags & TGL_UPDATE_WORKING || flags & TGL_UPDATE_DELETED) {
+    write_secret_chat_file (TLS);
+  }
+
+  PurpleBuddy *buddy = p2tgl_buddy_find (TLS, U->id);
+  
+  if (! (flags & TGL_UPDATE_DELETED)) {
+    if (!buddy) {
+      buddy = p2tgl_buddy_new (TLS, (tgl_peer_t *)U);
+      purple_blist_add_buddy (buddy, NULL, tggroup, NULL);
+      purple_blist_alias_buddy (buddy, U->print_name);
+    }
+    p2tgl_prpl_got_set_status_mobile (TLS, U->id);
+  }
+  
+  if (flags & TGL_UPDATE_REQUESTED && buddy)  {
+    connection_data *conn = TLS->ev_base;
+    
+    const char* choice = purple_account_get_string (conn->pa, "accept-secret-chats", "ask");
+    if (! strcmp (choice, "always")) {
+      tgl_do_accept_encr_chat_request (TLS, U, write_secret_chat_cb, 0);
+      
+    } else if (! strcmp(choice, "ask")) {
+      PurpleBuddy *who = p2tgl_buddy_find (TLS, TGL_MK_USER(U->user_id));
+      
+      struct accept_secret_chat_data *data = g_new (struct accept_secret_chat_data, 1);
+      data->TLS = TLS;
+      data->U = U;
+     
+      gchar *message = g_strdup_printf ("Accept Secret Chat '%s'?", U->print_name);
+      
+      purple_request_accept_cancel (conn->gc, "Secret Chat", message, "Secret chats can only have one "
+                                    "end point. If you accept a secret chat on this device, its messages will "
+                                    "not be available anywhere else. If you decline, you can accept"
+                                    " the chat on other devices.", 0, conn->pa, who->name, NULL, data,
+                                    G_CALLBACK(accept_secret_chat_cb), G_CALLBACK(decline_secret_chat_cb));
+      g_free (message);
+    }
+  }
+  
+  if (flags & TGL_UPDATE_CREATED && buddy) {
+    purple_buddy_set_protocol_data (buddy, (gpointer)U);
+    p2tgl_buddy_update (TLS, (tgl_peer_t *)U, flags);
+  }
+  
+  if (flags & TGL_UPDATE_DELETED && buddy) {
+    p2tgl_got_im (TLS, U->id, "Secret chat terminated.", PURPLE_MESSAGE_SYSTEM | PURPLE_MESSAGE_WHISPER, time(0));
+    p2tgl_prpl_got_set_status_offline (TLS, U->id);
+  }
+}
+
 static void update_chat_handler (struct tgl_state *TLS, struct tgl_chat *chat, unsigned flags) {
-  //tgl_do_get_chat_info (TLS, chat->id, 0, on_chat_get_info, 0);
   PurpleChat *ch = p2tgl_chat_find (TLS, chat->id);
   
   if (flags & TGL_UPDATE_CREATED) {
@@ -378,17 +574,14 @@ static void update_chat_handler (struct tgl_state *TLS, struct tgl_chat *chat, u
   if (flags & TGL_UPDATE_TITLE && ch) {
     purple_blist_alias_chat (ch, chat->print_title);
   }
-  if (flags & (TGL_UPDATE_MEMBERS | TGL_UPDATE_ADMIN)) {
+  if (flags & (TGL_UPDATE_MEMBERS | TGL_UPDATE_ADMIN) && ch) {
     on_update_chat_participants (TLS, chat);
   }
-  if (flags & TGL_UPDATE_DELETED) {
-    PurpleChat *ch = p2tgl_chat_find (TLS, chat->id);
-    if (ch) {
-      purple_blist_remove_chat (ch);
-    } else {
-      warning ("Cannot delete chat %d, not in buddy list.", chat->id);
-    }
+  if (flags & TGL_UPDATE_DELETED && ch) {
+    purple_blist_remove_chat (ch);
   }
+  
+  // TODO: check if user is a member of the current chat and don't display the chat in that case
 }
 
 static void update_user_typing (struct tgl_state *TLS, struct tgl_user *U, enum tgl_typing_status status) {
@@ -421,13 +614,15 @@ static void on_userpic_loaded (struct tgl_state *TLS, void *extra, int success,
     struct tgl_user *U = dld->data;
     warning ("Can not load userpic for user %s %s\n", U->first_name, U->last_name);
   }
-  telegram_conn *conn = TLS->ev_base;
+  connection_data *conn = TLS->ev_base;
 
   gchar *data = NULL;
   size_t len;
   GError *err = NULL;
   g_file_get_contents (filename, &data, &len, &err);
+  
   int imgStoreId = purple_imgstore_add_with_id (g_memdup(data, (guint)len), len, NULL);
+  used_images_add (conn, imgStoreId);
   
   struct download_desc *dld = extra;
   struct tgl_user *U = dld->data;
@@ -437,30 +632,72 @@ static void on_userpic_loaded (struct tgl_state *TLS, void *extra, int success,
   }
 
   char *who = g_strdup_printf ("%d", tgl_get_peer_id (U->id));
-  if (dld->type == 1) {
+  if (dld->get_user_info_data->show_info == 1) {
     PurpleNotifyUserInfo *info = create_user_notify_info(U);
-    char *profile_image = profile_image = format_img_full(imgStoreId);
+    
+    if (dld->get_user_info_data->peer && dld->get_user_info_data->peer->encr_chat.first_key_sha[0]) {
+      
+      // display secret key
+      int sha1key_store_id = generate_ident_icon (conn->TLS, dld->get_user_info_data->peer->encr_chat.first_key_sha);
+      if (sha1key_store_id != -1) {
+        char *ident_icon = format_img_full (sha1key_store_id);
+        purple_notify_user_info_add_pair (info, "Secret key", ident_icon);
+        g_free(ident_icon);
+      }
+      
+      g_free (who);
+      who = g_strdup_printf ("%d", tgl_get_peer_id (dld->get_user_info_data->peer->id));
+    }
+    
+    char *profile_image = profile_image = format_img_full (imgStoreId);
     purple_notify_user_info_add_pair (info, "Profile image", profile_image);
     purple_notify_userinfo (conn->gc, who, info, NULL, NULL);
+    
     g_free (profile_image);
   }
-  purple_buddy_icons_set_for_user(conn->pa, who, g_memdup(data, (guint)len), len, NULL);
+  if (dld->get_user_info_data->peer) {
+     char *id = g_strdup_printf ("%d", tgl_get_peer_id (dld->get_user_info_data->peer->id));
+     gchar *img_data = NULL;
+     size_t len;
+     GError *err = NULL;
+     g_file_get_contents (filename, &img_data, &len, &err);
+     purple_buddy_icons_set_for_user(conn->pa, id, img_data, len, NULL);
+     g_free (id);
+  }
+  purple_buddy_icons_set_for_user(conn->pa, who, data, len, NULL);
   g_free(who);
+  g_free(dld->get_user_info_data);
 }
 
-void on_user_get_info (struct tgl_state *TLS, void *show_info, int success, struct tgl_user *U)
+void on_user_get_info (struct tgl_state *TLS, void *info_data, int success, struct tgl_user *U)
 {
-  assert (success);
+  if (! success) {
+    warning ("on_user_get_info not successfull, aborting...\n");
+    return;
+  }
   
+  get_user_info_data *user_info_data = (get_user_info_data *)info_data;
   if (U->photo.sizes_num == 0) {
-    if (show_info) {
+    if (user_info_data->show_info) {
       PurpleNotifyUserInfo *info = create_user_notify_info(U);
-      p2tgl_notify_userinfo(TLS, U->id, info, NULL, NULL);
+      if (user_info_data->peer && user_info_data->peer->encr_chat.first_key_sha[0]) {
+        // display secret key
+        int sha1key_store_id = generate_ident_icon (TLS, user_info_data->peer->encr_chat.first_key_sha);
+        if (sha1key_store_id != -1) {
+          char *ident_icon = format_img_full (sha1key_store_id);
+          purple_notify_user_info_add_pair (info, "Secret key", ident_icon);
+          g_free (ident_icon);
+        }
+        p2tgl_notify_userinfo (TLS, user_info_data->peer->id, info, NULL, NULL);
+      } else {
+	      p2tgl_notify_userinfo (TLS, U->id, info, NULL, NULL);
+      }
     }
+    g_free(user_info_data);
   } else {
     struct download_desc *dld = malloc (sizeof(struct download_desc));
     dld->data = U;
-    dld->type = show_info ? 1 : 2;
+    dld->get_user_info_data = info_data;
     tgl_do_load_photo (TLS, &U->photo, on_userpic_loaded, dld);
   }
 }
@@ -469,15 +706,15 @@ void on_chat_get_info (struct tgl_state *TLS, void *extra, int success, struct t
   assert (success);
   
   debug ("on_chat_joined(%d)\n", tgl_get_peer_id (C->id));
-  telegram_conn *conn = TLS->ev_base;
+  connection_data *conn = TLS->ev_base;
 
   PurpleConversation *conv;
-  if (!(conv = purple_find_chat(conn->gc, tgl_get_peer_id(C->id)))) {
+  if (!(conv = purple_find_chat (conn->gc, tgl_get_peer_id(C->id)))) {
     // chat conversation is not existing, create it
-    conv = serv_got_joined_chat(conn->gc, tgl_get_peer_id(C->id), C->title);
+    conv = serv_got_joined_chat (conn->gc, tgl_get_peer_id(C->id), C->title);
   }
-  purple_conv_chat_clear_users(purple_conversation_get_chat_data(conv));
-  chat_add_all_users(conv, C);
+  purple_conv_chat_clear_users (purple_conversation_get_chat_data(conv));
+  chat_add_all_users (conv, C);
   
   struct message_text *mt = 0;
   while ((mt = g_queue_pop_head (conn->new_messages))) {
@@ -491,21 +728,20 @@ void on_chat_get_info (struct tgl_state *TLS, void *extra, int success, struct t
     free (mt);
   }
   
-  gchar *name = g_strdup_printf ("%d", tgl_get_peer_id(C->id));
+  gchar *name = g_strdup_printf ("%d", tgl_get_peer_id (C->id));
   g_hash_table_remove (conn->joining_chats, name);
   g_free (name);
 }
 
 void on_ready (struct tgl_state *TLS) {
   debug ("on_ready().\n");
-  telegram_conn *conn = TLS->ev_base;
+  connection_data *conn = TLS->ev_base;
   
   purple_connection_set_state(conn->gc, PURPLE_CONNECTED);
   purple_connection_set_display_name(conn->gc, purple_account_get_username(conn->pa));
   purple_blist_add_account(conn->pa);
-  tggroup = purple_find_group("Telegram");
+  tggroup = purple_find_group ("Telegram");
   if (tggroup == NULL) {
-    debug ("PurpleGroup = NULL, creating");
     tggroup = purple_group_new ("Telegram");
     purple_blist_add_group (tggroup, NULL);
   }
@@ -543,20 +779,43 @@ static GList *tgprpl_status_types (PurpleAccount * acct) {
   debug ("tgprpl_status_types()\n");
   GList *types = NULL;
   PurpleStatusType *type;
-  type = purple_status_type_new_with_attrs (PURPLE_STATUS_AVAILABLE, NULL, NULL,
-          1, 1, 0, "last online", "last online", purple_value_new (PURPLE_TYPE_STRING), NULL);
+  
+  type = purple_status_type_new_full(PURPLE_STATUS_AVAILABLE, NULL, NULL, FALSE, TRUE, FALSE);
   types = g_list_prepend (types, type);
   
-  type = purple_status_type_new_with_attrs (PURPLE_STATUS_MOBILE, NULL, NULL, 1,
-          1, 0, "last online", "last online", purple_value_new (PURPLE_TYPE_STRING), NULL);
+  type = purple_status_type_new_full(PURPLE_STATUS_MOBILE, NULL, NULL, FALSE, TRUE, FALSE);
+  types = g_list_prepend (types, type);
+
+  type = purple_status_type_new_full(PURPLE_STATUS_OFFLINE, NULL, NULL, FALSE, TRUE, FALSE);
   types = g_list_prepend (types, type);
   
-  type = purple_status_type_new (PURPLE_STATUS_OFFLINE, NULL, NULL, 1);
-  types = g_list_append (types, type);
+  /*
+    The states below are only registered internally so that we get notified about 
+    state changes to away and unavailable. This is useful for deciding when to send 
+    No other peer should ever have those states.
+   */
+  type = purple_status_type_new_full(PURPLE_STATUS_AWAY, NULL, NULL, FALSE, TRUE, FALSE);
+  types = g_list_prepend (types, type);
+  type = purple_status_type_new_full(PURPLE_STATUS_UNAVAILABLE, NULL, NULL, FALSE, TRUE, FALSE);
+  types = g_list_prepend (types, type);
   
   return g_list_reverse (types);
 }
 
+static GList* tgprpl_blist_node_menu (PurpleBlistNode *node) {
+  purple_debug_info (PLUGIN_ID, "tgprpl_blist_node_menu()\n");
+  
+  GList* menu = NULL;
+  if (PURPLE_BLIST_NODE_IS_BUDDY(node)) {
+    // Add encrypted chat option to the right click menu of all buddies
+    PurpleBuddy* buddy = (PurpleBuddy*)node;
+    PurpleMenuAction* menu_action = purple_menu_action_new("Start Secret Chat",
+                                      PURPLE_CALLBACK(start_secret_chat), buddy, NULL);
+    menu = g_list_append(menu, (gpointer)menu_action);
+  }
+  return menu;
+}
+
 static GList *tgprpl_chat_join_info (PurpleConnection * gc) {
   debug ("tgprpl_chat_join_info()\n");
   struct proto_chat_entry *pce;
@@ -595,12 +854,7 @@ static void tgprpl_login (PurpleAccount * acct) {
   
   // create handle to store additional info for libpurple in
   // the new telegram instance
-  telegram_conn *conn = g_new0(telegram_conn, 1);
-  conn->TLS = TLS;
-  conn->gc = gc;
-  conn->pa = acct;
-  conn->new_messages = g_queue_new ();
-  conn->joining_chats = g_hash_table_new_full(g_str_hash, g_str_equal, g_free, g_free);
+  connection_data *conn = connection_data_init (TLS, gc, acct);
   purple_connection_set_protocol_data (gc, conn);
   
   tgl_set_ev_base (TLS, conn);
@@ -617,38 +871,49 @@ static void tgprpl_login (PurpleAccount * acct) {
 
 static void tgprpl_close (PurpleConnection * gc) {
   debug ("tgprpl_close()\n");
-  telegram_conn *conn = purple_connection_get_protocol_data(gc);
-  purple_timeout_remove(conn->timer);
-  tgl_free_all (conn->TLS);
+  connection_data *conn = purple_connection_get_protocol_data (gc);
+  purple_timeout_remove (conn->timer);
+  
+  connection_data_free (conn);
 }
 
 static int tgprpl_send_im (PurpleConnection * gc, const char *who, const char *message, PurpleMessageFlags flags) {
   debug ("tgprpl_send_im()\n");
-  
-  telegram_conn *conn = purple_connection_get_protocol_data(gc);
-  PurpleAccount *pa = conn->pa;
-  
-  PurpleBuddy *b = purple_find_buddy (pa, who);
-  if (!b) {
-    warning ("Buddy %s not found, cannot send IM\n", who);
-    return -1;
+  connection_data *conn = purple_connection_get_protocol_data(gc);
+ 
+  // this is part of a workaround to support clients without
+  // the request API (request.h), see telegram-base.c:request_code()
+  if (conn->in_fallback_chat) {
+    request_code_entered (conn->TLS, message);
+    conn->in_fallback_chat = 0;
+    return 1;
   }
   
-  tgl_peer_t *peer = tgl_peer_get(conn->TLS, TGL_MK_USER(atoi (who)));
-  if (!peer) {
-    warning ("Protocol data tgl_peer_t for %s not found, cannot send IM\n", who);
-    return -1;
+  /* 
+     Make sure that we only send messages to an existing peer by
+     searching it in the peer tree. This allows us to give immediate feedback 
+     by returning an error-code in case the peer doesn't exist
+   */
+  tgl_peer_t *peer = find_peer_by_name (conn->TLS, who);
+  if (peer) {
+
+    if (tgl_get_peer_type(peer->id) == TGL_PEER_ENCR_CHAT && peer->encr_chat.state != sc_ok) {
+      // secret chat not ready for sending messages or deleted
+      return -1;
+    }
+    
+    tgl_do_send_unescape_message (conn->TLS, message, peer->id);
+    return 1;
   }
-  gchar *raw = purple_unescape_html(message);
-  tgl_do_send_message (conn->TLS, peer->id, raw, (int)strlen (raw), 0, 0);
-  g_free(raw);
-  return 1;
+  
+  // peer not found
+  return -1;
 }
 
 static unsigned int tgprpl_send_typing (PurpleConnection * gc, const char *who, PurpleTypingState typing) {
   debug ("tgprpl_send_typing()\n");
   int id = atoi (who);
-  telegram_conn *conn = purple_connection_get_protocol_data(gc);
+  connection_data *conn = purple_connection_get_protocol_data(gc);
   tgl_peer_t *U = tgl_peer_get (conn->TLS, TGL_MK_USER (id));
   if (U) {
     if (typing == PURPLE_TYPING) {
@@ -660,41 +925,90 @@ static unsigned int tgprpl_send_typing (PurpleConnection * gc, const char *who,
   return 0;
 }
 
-static void tgprpl_get_info (PurpleConnection * gc, const char *username) {
+static void tgprpl_get_info (PurpleConnection * gc, const char *who) {
   debug ("tgprpl_get_info()\n");
-  telegram_conn *conn = purple_connection_get_protocol_data(gc);
-  tgl_peer_id_t u = TGL_MK_USER(atoi(username));
-  tgl_do_get_user_info (conn->TLS, u, 0, on_user_get_info, (void *)1l);
+  connection_data *conn = purple_connection_get_protocol_data(gc);
+  
+  tgl_peer_t *peer = find_peer_by_name (conn->TLS, who);
+  if (! peer) { return; }
+  
+  get_user_info_data* info_data = malloc (sizeof(get_user_info_data));
+  info_data->show_info = 1;
+  info_data->peer = peer;
+  
+  switch (tgl_get_peer_type (peer->id)) {
+    case TGL_PEER_USER:
+    case TGL_PEER_CHAT:
+      tgl_do_get_user_info (conn->TLS, peer->id, 0, on_user_get_info, info_data);
+      break;
+    case TGL_PEER_ENCR_CHAT: {
+      tgl_peer_t *parent_peer;
+      
+      if (peer->encr_chat.state == sc_waiting) {
+        PurpleNotifyUserInfo *info = purple_notify_user_info_new();
+        purple_notify_user_info_add_pair (info, "", "Waiting for user to get online ...");
+        p2tgl_notify_userinfo (conn->TLS, peer->id, info, NULL, NULL);
+        return;
+      }
+      
+      if (peer->encr_chat.admin_id == conn->TLS->our_id){
+        parent_peer = tgl_peer_get (conn->TLS, TGL_MK_USER(peer->encr_chat.user_id));
+      } else {
+        parent_peer = tgl_peer_get (conn->TLS, TGL_MK_USER(peer->encr_chat.admin_id));
+      }
+      if (parent_peer){
+        tgl_do_get_user_info(conn->TLS, parent_peer->id, 0, on_user_get_info, info_data);
+      }
+      break;
+    }
+      
+  }
 }
 
 static void tgprpl_set_status (PurpleAccount * acct, PurpleStatus * status) {
-  debug ("tgprpl_set_status()\n");
+  debug ("tgprpl_set_status(%s)\n", purple_status_get_name (status));
+  debug ("tgprpl_set_status(currstatus=%s)\n", purple_status_get_name(purple_account_get_active_status(acct)));
+
+  PurpleConnection *gc = purple_account_get_connection(acct);
+  if (!gc) { return; }
+  connection_data *conn = purple_connection_get_protocol_data (gc);
+
+  if (p2tgl_status_is_present(status)) {
+    pending_reads_send_all (conn->pending_reads, conn->TLS);
+  }
 }
 
 static void tgprpl_add_buddy (PurpleConnection * gc, PurpleBuddy * buddy, PurpleGroup * group) {
-  telegram_conn *conn = purple_connection_get_protocol_data(gc);
+  connection_data *conn = purple_connection_get_protocol_data(gc);
   const char* first = buddy->alias ? buddy->alias : "";
   
   tgl_do_add_contact (conn->TLS, buddy->name, (int)strlen (buddy->name), first, (int)strlen (first), "", 0, 0, on_contact_added, buddy);
 }
 
-static void tgprpl_add_buddies (PurpleConnection * gc, GList * buddies, GList * groups) {
-  debug ("tgprpl_add_buddies()\n");
-}
-
 static void tgprpl_remove_buddy (PurpleConnection * gc, PurpleBuddy * buddy, PurpleGroup * group) {
   debug ("tgprpl_remove_buddy()\n");
   if (!buddy) { return; }
-  
-  telegram_conn *conn = purple_connection_get_protocol_data (gc);
-  struct tgl_user *user = purple_buddy_get_protocol_data (buddy);
-  if (!user) { warning ("cannot remove buddy '%s', no protocol data found\n", buddy->name); return; }
-  
-  tgl_do_del_contact (conn->TLS, user->id, NULL, NULL);
-}
 
-static void tgprpl_remove_buddies (PurpleConnection * gc, GList * buddies, GList * groups){
-  debug ("tgprpl_remove_buddies()\n");
+  connection_data *conn = purple_connection_get_protocol_data (gc);
+  
+  tgl_peer_t *peer = find_peer_by_name (conn->TLS, buddy->name);
+  if (!peer) {
+    // telegram peer not existing, orphaned buddy
+    return;
+  }
+  
+  switch (tgl_get_peer_type(peer->id)) {
+    case TGL_PEER_ENCR_CHAT:
+      /* TODO: implement the api call cancel secret chats. Currently the chat will only be marked as 
+               deleted on our side so that it won't be added on startup 
+              (when the secret chat file is loaded) */
+      bl_do_encr_chat_delete (conn->TLS, &peer->encr_chat);
+      break;
+    case TGL_PEER_USER:
+      tgl_do_del_contact (conn->TLS, peer->id, NULL, NULL);
+      break;
+  }
+  
 }
 
 static void tgprpl_add_deny (PurpleConnection * gc, const char *name){
@@ -708,7 +1022,7 @@ static void tgprpl_rem_deny (PurpleConnection * gc, const char *name){
 static void tgprpl_chat_join (PurpleConnection * gc, GHashTable * data) {
   debug ("tgprpl_chat_join()\n");
   
-  telegram_conn *conn = purple_connection_get_protocol_data (gc);
+  connection_data *conn = purple_connection_get_protocol_data (gc);
   const char *groupname = g_hash_table_lookup (data, "subject");
   
   char *id = g_hash_table_lookup(data, "id");
@@ -745,7 +1059,7 @@ static GHashTable *tgprpl_chat_info_deflt (PurpleConnection * gc, const char *ch
 static void tgprpl_chat_invite (PurpleConnection * gc, int id, const char *message, const char *name) {
   debug ("tgprpl_chat_invite()\n");
 
-  telegram_conn *conn = purple_connection_get_protocol_data (gc);
+  connection_data *conn = purple_connection_get_protocol_data (gc);
   tgl_peer_t *chat = tgl_peer_get(conn->TLS, TGL_MK_CHAT (id));
   tgl_peer_t *user = tgl_peer_get(conn->TLS, TGL_MK_USER (atoi(name)));
   
@@ -759,14 +1073,17 @@ static void tgprpl_chat_invite (PurpleConnection * gc, int id, const char *messa
 
 static int tgprpl_send_chat (PurpleConnection * gc, int id, const char *message, PurpleMessageFlags flags) {
   debug ("tgprpl_send_chat()\n");
-  telegram_conn *conn = purple_connection_get_protocol_data (gc);
-  
-  gchar *raw = purple_unescape_html(message);
-  tgl_do_send_message (conn->TLS, TGL_MK_CHAT(id), raw, (int)strlen (raw), 0, 0);
-  g_free (raw);
+  connection_data *conn = purple_connection_get_protocol_data (gc);
+  tgl_do_send_unescape_message (conn->TLS, message, TGL_MK_CHAT(id));
   
+  /* Pidgin won't display the written message if we don't call this, Adium will display it twice 
+     if we call it, so we don't do it for the adium Plugin.
+     TODO: there has to be a better way to do this, figure out how.
+   */
+#ifndef __ADIUM_
   p2tgl_got_chat_in(conn->TLS, TGL_MK_CHAT(id), TGL_MK_USER(conn->TLS->our_id), message,
-                    PURPLE_MESSAGE_RECV, time(NULL));
+    PURPLE_MESSAGE_RECV, time(NULL));
+#endif
   return 1;
 }
 
@@ -785,7 +1102,7 @@ static void tgprpl_convo_closed (PurpleConnection * gc, const char *who){
 static void tgprpl_set_buddy_icon (PurpleConnection * gc, PurpleStoredImage * img) {
   debug ("tgprpl_set_buddy_icon()\n");
   
-  telegram_conn *conn = purple_connection_get_protocol_data (gc);
+  connection_data *conn = purple_connection_get_protocol_data (gc);
   if (purple_imgstore_get_filename (img)) {
     char* filename = g_strdup_printf ("%s/icons/%s", purple_user_dir(), purple_imgstore_get_filename (img));
     debug (filename);
@@ -824,7 +1141,7 @@ static PurplePluginProtocolInfo prpl_info = {
   NULL,
   tgprpl_tooltip_text,
   tgprpl_status_types,
-  NULL,                    // blist_node_menu
+  tgprpl_blist_node_menu,  // blist_node_menu
   tgprpl_chat_join_info,
   tgprpl_chat_info_deflt,
   tgprpl_login,
@@ -837,9 +1154,9 @@ static PurplePluginProtocolInfo prpl_info = {
   NULL,                    // set_idle
   NULL,                    // change_passwd
   tgprpl_add_buddy,
-  tgprpl_add_buddies,
+  NULL,                    // add_buddies
   tgprpl_remove_buddy,
-  tgprpl_remove_buddies,
+  NULL,                    // remove_buddies
   NULL,                    // add_permit
   tgprpl_add_deny,
   NULL,                    // rem_permit
@@ -891,18 +1208,27 @@ static PurplePluginProtocolInfo prpl_info = {
   NULL                     // add_buddies_with_invite
 };
 
-
 static void tgprpl_init (PurplePlugin *plugin) {
-  //PurpleAccountOption *option;
-  //GList *verification_values = NULL;
   
   PurpleAccountOption *opt;
   
-  opt = purple_account_option_bool_new("Compatibility Mode (read SMS code from settings)",
-          "compat-verification", 0);
+  GList *verification_values = NULL;
+  
+#define ADD_VALUE(list, desc, v) { \
+  PurpleKeyValuePair *kvp = g_new0(PurpleKeyValuePair, 1); \
+  kvp->key = g_strdup((desc)); \
+  kvp->value = g_strdup((v)); \
+  list = g_list_prepend(list, kvp); \
+}
+    
+  ADD_VALUE(verification_values, "Ask", "ask");
+  ADD_VALUE(verification_values, "Always", "always");
+  ADD_VALUE(verification_values, "Never", "never");
+ 
+  opt = purple_account_option_list_new("Accept Secret Chats", "accept-secret-chats", verification_values);
   prpl_info.protocol_options = g_list_append(prpl_info.protocol_options, opt);
   
-  opt = purple_account_option_string_new("SMS Code", "code", "");
+  opt = purple_account_option_bool_new("Fallback SMS verification", "compat-verification", 0);
   prpl_info.protocol_options = g_list_append(prpl_info.protocol_options, opt);
   
   _telegram_protocol = plugin;
